#include <iostream>
#include <memory>
#include <vector>

struct animal {
	virtual void speak() const {
		std::cout << "<generic animal noise>\n";
	}

	virtual ~animal() = default;
};

struct dog : animal {
	void speak() const override {
		std::cout << "Whoof!\n";
	}
};

struct cat : animal {
	void speak() const override {
		std::cout << "Meow!\n";
	}

	void knock_things_down() const {
		std::cout << "Oops… Crash!\n";
	}
};

struct bad_wizard {
	animal familiar;
};

struct good_wizard {
	std::unique_ptr<animal> familiar;
};

int main() {
	cat felix;
	felix.speak();

	animal& ref = felix;
	ref.speak();

	// use clang-tidy
	animal pet = felix;
	pet.speak();

	std::cout << "---\n";

	bad_wizard rincewind = { cat() };
	rincewind.familiar.speak();

	good_wizard ridcully = { std::make_unique<cat>() };
	ridcully.familiar->speak();

	std::cout << "---\n";

	std::vector<animal> animals;
	animals.push_back(cat());
	animals.push_back(dog());

	for (const auto& animal : animals) {
		animal.speak();
	}

	std::cout << "---\n";

	std::vector<std::unique_ptr<animal>> animals2;
	animals2.push_back(std::make_unique<cat>());
	animals2.push_back(std::make_unique<dog>());

	for (const auto& ptr : animals2) {
		ptr->speak();
		if (auto* cat_ptr = dynamic_cast<cat*>(ptr.get())) {
			cat_ptr->knock_things_down();
		}
	}

	std::cout << "---\n";

	auto& cat_ref = dynamic_cast<cat&>(ref);
	cat_ref.knock_things_down();
}
