#include <iostream>
#include <memory>
#include <vector>

struct animal {
	virtual void speak() const {
		std::cout << "<generic animal noise>\n";
	}

	virtual std::unique_ptr<animal> clone() const {
		return std::make_unique<animal>(*this);
	}

	virtual ~animal() = default;
};

struct dog : animal {
	void speak() const override {
		std::cout << "Whoof!\n";
	}
	std::unique_ptr<animal> clone() const override {
		return std::make_unique<dog>(*this);
	}
};

struct cat : animal {
	void speak() const override {
		std::cout << "Meow!\n";
	}
	std::unique_ptr<animal> clone() const override {
		return std::make_unique<cat>(*this);
	}
};

int main() {
	cat felix;
	animal& pet = felix;

	auto clone = pet.clone();

	std::vector<std::unique_ptr<animal>> animals;
	animals.push_back(std::make_unique<dog>());
	animals.push_back(std::make_unique<cat>());

	std::vector<std::unique_ptr<animal>> clones;
	for (const auto& ptr : animals) {
		clones.push_back(ptr->clone());
	}
}
